# vim: expandtab:ts=4:sw=4
import numpy as np
from .kalman_filter import KalmanFilter


class TrackState:
    """
    Enumeration type for the single target track state. Newly created tracks are
    classified as `tentative` until enough evidence has been collected. Then,
    the track state is changed to `confirmed`. Tracks that are no longer alive
    are classified as `deleted` to mark them for removal from the set of active
    tracks.

    """

    Tentative = 't'
    Confirmed = 'c'
    Deleted = 'd'


class Track:
    """
    A single target track with state space `(x, y, a, h)` and associated
    velocities, where `(x, y)` is the center of the bounding box, `a` is the
    aspect ratio (w/h) and `h` is the height.

    Parameters
    ----------
    mean : ndarray
        Mean vector of the initial state distribution.
    covariance : ndarray
        Covariance matrix of the initial state distribution.
    track_id : int
        A unique track identifier.
    n_init : int
        Number of consecutive detections before the track is confirmed. The
        track state is set to `Deleted` if a miss occurs within the first
        `n_init` frames.
    max_age : int
        The maximum number of consecutive misses before the track state is
        set to `Deleted`.
    feature : Optional[ndarray]
        Feature vector of the detection this track originates from. If not None,
        this feature is added to the `features` cache.

    Attributes
    ----------
    mean : ndarray
        Mean vector of the initial state distribution.
    covariance : ndarray
        Covariance matrix of the initial state distribution.
    track_id : int
        A unique track identifier.
    hits : int
        Total number of measurement updates.
    age : int
        Total number of frames since first occurance.
    time_since_update : int
        Total number of frames since last measurement update.
    state : TrackState
        The current track state.
    features : List[ndarray]
        A cache of features. On each measurement update, the associated feature
        vector is added to this list.

    """

    def __init__(self, detection, track_id, class_id, conf, n_init, max_age, ema_alpha, max_kalman_prediction_without_update,
                 feature=None):
        self.track_id = track_id
        self.class_id = int(class_id)
        self.hits = 1
        self.age = 1
        self.time_since_update = 0
        self.ema_alpha = ema_alpha

        self.state = TrackState.Tentative
        self.features = []
        if feature is not None:
            # feature["reid_features"] /= np.linalg.norm(feature["reid_features"])  # TODO can cause div by 0 + check if norm is already performed in compute_distance_matrix_using_bp_features
            self.features.append(feature)

        self.conf = conf
        self._n_init = n_init
        self._max_age = max_age

        self.kf = KalmanFilter()
        self.mean, self.covariance = self.kf.initiate(detection.to_xyah())
        self.last_detection = detection
        self.update_state()  # update state to confirmed if n_init = 1
        self.last_kf_pred_ltwh = None
        self.max_kalman_prediction_without_update = max_kalman_prediction_without_update

    def to_xyah(self):
        return self.mean[:4].copy()

    def to_ltwh(self):
        xyah = self.to_xyah()
        width = xyah[2] * xyah[3]
        return np.array([xyah[0] - width / 2, xyah[1] - xyah[3] / 2, width, xyah[3]])

    def to_tlbr(self):
        """Get kf estimated current position in bounding box format `(min x, miny, max x,
        max y)`.

        Returns
        -------
        ndarray
            The predicted kf bounding box.

        """
        ret = self.to_ltwh()
        ret[2:] = ret[:2] + ret[2:]
        return ret

    def camera_update(self, matrix):
        x1, y1, x2, y2 = self.to_tlbr()
        x1_, y1_, _ = matrix @ np.array([x1, y1, 1]).T
        x2_, y2_, _ = matrix @ np.array([x2, y2, 1]).T
        w, h = x2_ - x1_, y2_ - y1_
        cx, cy = x1_ + w / 2, y1_ + h / 2
        self.mean[:4] = [cx, cy, w/h, h]

    def increment_age(self):
        self.age += 1
        self.time_since_update += 1

    def predict(self):
        """Propagate the state distribution to the current time step using a
        Kalman filter prediction step.
        """
        if self.time_since_update < self.max_kalman_prediction_without_update:
            self.mean, self.covariance = self.kf.predict(self.mean, self.covariance)
        self.age += 1
        self.time_since_update += 1

    def update(self, detection, class_id, conf):
        """Perform Kalman filter measurement update step and update the feature
        cache.
        Parameters
        ----------
        detection : Detection
            The associated detection.
        """
        self.conf = conf
        self.class_id = class_id.int()
        self.last_detection = detection
        self.last_kf_pred_ltwh = self.to_ltwh()
        self.mean, self.covariance = self.kf.update(self.mean, self.covariance, detection.to_xyah(), detection.confidence)

        detection_features = detection.feature['reid_features']
        detection_vis_scores = detection.feature['visibility_scores']

        tracklet_features = self.features[-1]['reid_features']
        tracklet_vis_scores = self.features[-1]['visibility_scores']

        xor = np.logical_xor(tracklet_vis_scores, detection_vis_scores)
        ema_scores_tracklet = (tracklet_vis_scores * detection_vis_scores) * np.float32(
            self.ema_alpha) + xor * tracklet_vis_scores
        ema_scores_detection = (tracklet_vis_scores * detection_vis_scores) * np.float32(
            1 - self.ema_alpha) + xor * detection_vis_scores
        smooth_feat = np.expand_dims(ema_scores_tracklet, 1) * tracklet_features + np.expand_dims(
            ema_scores_detection, 1) * detection_features
        smooth_visibility_scores = np.maximum(tracklet_vis_scores, detection_vis_scores)
        smooth_feat[np.logical_and(ema_scores_tracklet == 0., ema_scores_detection == 0.)] = 1
        # smooth_feat /= np.linalg.norm(smooth_feat, axis=-1, keepdims=True)  # TODO can cause div by 0 + check if norm is already performed in compute_distance_matrix_using_bp_features
        self.features = [{
            'reid_features': smooth_feat,
            'visibility_scores': smooth_visibility_scores
        }]

        self.hits += 1
        self.time_since_update = 0
        self.update_state()

    def update_state(self):
        """Update the state of the tracklet to "confirmed" if it was updated enough times"""
        if self.state == TrackState.Tentative and self.hits >= self._n_init:
            self.state = TrackState.Confirmed

    def mark_missed(self):
        """Mark this track as missed (no association at the current time step).
        """
        if self.state == TrackState.Tentative:
            self.state = TrackState.Deleted
        elif self.time_since_update > self._max_age:
            self.state = TrackState.Deleted

    def is_tentative(self):
        """Returns True if this track is tentative (unconfirmed).
        """
        return self.state == TrackState.Tentative

    def is_confirmed(self):
        """Returns True if this track is confirmed."""
        return self.state == TrackState.Confirmed

    def is_deleted(self):
        """Returns True if this track is dead and should be deleted."""
        return self.state == TrackState.Deleted
